--------------------------------------------------------------------------------
--
-- File: shuriken.vhd
-- Version: 1.1 Start date 8/9/2014 (Steven Taffs)
-- Version: 1.2 Start date 13/11/2016 (Steven Taffs)
--
-- Description: 
--
-- Loads OS from serial flash onto flipper exi data in line
--
-- Targeted device: CPLD  xc9572xl-10vq44
-- Author: Steven Taffs
--
--------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.std_logic_unsigned.all;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
--use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

----------------------------------------------------------------
-- Main entity
----------------------------------------------------------------
entity shuriken is
    Port ( 
	        --gamecube connections
	        exi_serial_in : in  STD_LOGIC;
           exi_serial_to_gc : out  STD_LOGIC;
           exi_clk : in  STD_LOGIC;
           exi_cs : in  STD_LOGIC;
			  --serial flash connections
			  serial_flash_in : out  STD_LOGIC;
			  serial_flash_out : in  STD_LOGIC;
			  serial_flash_cs : out  STD_LOGIC;
           --PCB switch option
			  bypass : in  STD_LOGIC
         );			  
end shuriken;

architecture arch_shuriken of shuriken is

----------------------------------------------------------------
-- signal defines
----------------------------------------------------------------

signal bit_counter: std_logic_vector(8 downto 0);    --counts number of bits sent
signal serial_payload_enable: std_logic;             --send payload to gc if this gets set
signal serial_payload_disable: std_logic := '0';     --stop payload to gc if this gets set

--serial flash read address counter (high / middle / low bytes)
signal serial_addr_counter: std_logic_vector(14 downto 0) := (others => '0');
--enables serial flash chip select line
signal serial_enable: std_logic := '0';

----------------------------------------------------------------
-- asynchronous events
----------------------------------------------------------------	
begin
   --control data to gc 
	exi_serial_to_gc <= ((serial_flash_out and serial_payload_enable and (not bypass)) or (exi_serial_in and (not serial_payload_enable) and (not bypass))
	                    or (exi_serial_in and bypass)) when (exi_cs = '0') else '0';
	
	--control serial flash chip select line
	serial_flash_cs <= exi_cs when (serial_enable = '1') else 'Z';
	
----------------------------------------------------------------
-- synchronised processes 
----------------------------------------------------------------					
   process (exi_cs, exi_clk, bypass)
		begin
		if ((exi_cs = '1') or (bypass = '1')) then
			bit_counter <= (others => '0');
			serial_payload_enable <= '0';
			serial_flash_in <= 'Z';
		else
			if (exi_clk'event and exi_clk='0' and serial_payload_disable = '0') then
			
				if(serial_addr_counter = "000000000000000") then    -- first transfer start from byte 32
			
					case bit_counter is				
						when "011111111" =>                        -- time to select serial flash chip
							 serial_enable <= '1';		         		
							 
						when "100000101" =>
							 serial_flash_in <= '1';                -- M25P40 serial flash chips read op-code "00000011"
							 
						when "100000110" =>
							 serial_flash_in <= '1';		
							 
						when "100001011" =>	                      -- send high / middle / low address counter value
							 serial_flash_in <= serial_addr_counter(14);
							 
						when "100001100" =>	                   
							 serial_flash_in <= serial_addr_counter(13);
							 
						when "100001101" =>	                   
							 serial_flash_in <= serial_addr_counter(12);
							 
						when "100001110" =>	                    
							 serial_flash_in <= serial_addr_counter(11);
							 
						when "100001111" =>	                    
							 serial_flash_in <= serial_addr_counter(10);
							 
						when "100010000" =>	                   
							 serial_flash_in <= serial_addr_counter(9);
							 
						when "100010001" =>	                   
							 serial_flash_in <= serial_addr_counter(8);
							 
						when "100010010" =>	                   
							 serial_flash_in <= serial_addr_counter(7);
							 
						when "100010011" =>	                   
							 serial_flash_in <= serial_addr_counter(6);
							 
						when "100010100" =>	                   
							 serial_flash_in <= serial_addr_counter(5);
							 
						when "100010101" =>	                   
							 serial_flash_in <= serial_addr_counter(4);
							 
						when "100010110" =>	                   
							 serial_flash_in <= serial_addr_counter(3);

                  when "100010111" =>	                   
							 serial_flash_in <= serial_addr_counter(2);	

						when "100011000" =>	                   
							 serial_flash_in <= serial_addr_counter(1);	

						when "100011001" =>	                   
							 serial_flash_in <= serial_addr_counter(0);										 
												
						when "100011111" =>                 -- time to send payload (switch to serial flash output)
							 serial_payload_enable <= '1';
							 serial_addr_counter <= serial_addr_counter + "11111"; -- inc address counter (1024 - 32)  			 
							 
						when others =>
							 serial_flash_in <= '0';         -- set read address lower 5 bits to zero

					end case;
				
				else  -- all other transfers start at byte 4
				
				   case bit_counter is											 							 
						when "000000101" =>
							 serial_flash_in <= '1';           -- M25P40 serial flash chips read op-code "00000011"
							 
						when "000000110" =>
							 serial_flash_in <= '1';
							 
						when "000001011" =>	                 -- send high / middle / low address counter value
							 serial_flash_in <= serial_addr_counter(14);
							 
						when "000001100" =>	                   
							 serial_flash_in <= serial_addr_counter(13);
							 
						when "000001101" =>	                   
							 serial_flash_in <= serial_addr_counter(12);
							 
						when "000001110" =>	                    
							 serial_flash_in <= serial_addr_counter(11);
							 
						when "000001111" =>	                    
							 serial_flash_in <= serial_addr_counter(10);
							 
						when "000010000" =>	                   
							 serial_flash_in <= serial_addr_counter(9);
							 
						when "000010001" =>	                   
							 serial_flash_in <= serial_addr_counter(8);
							 
						when "000010010" =>	                   
							 serial_flash_in <= serial_addr_counter(7);
							 
						when "000010011" =>	                   
							 serial_flash_in <= serial_addr_counter(6);
							 
						when "000010100" =>	                   
							 serial_flash_in <= serial_addr_counter(5);
							 
						when "000010101" =>	                   
							 serial_flash_in <= serial_addr_counter(4);
							 
						when "000010110" =>	                   
							 serial_flash_in <= serial_addr_counter(3);

                  when "000010111" =>	                   
							 serial_flash_in <= serial_addr_counter(2);	

						when "000011000" =>	                   
							 serial_flash_in <= serial_addr_counter(1);	

						when "000011001" =>	                   
							 serial_flash_in <= serial_addr_counter(0);							 
												
						when "000011111" =>                 -- time to send payload (switch to serial flash output)
							 serial_payload_enable <= '1';
							 serial_addr_counter <= serial_addr_counter + "100000"; 	-- inc address counter (1024)  
							 
						when others =>
							 serial_flash_in <= '0';         -- set read address lower 5 bits to zero

					end case;
				
				end if;
				
				if(bit_counter < "111111111") then 
				    bit_counter <= bit_counter + '1';     -- inc bit counter
			   end if;
				
				if(serial_addr_counter = "110000000011111") then  -- check if the whole 768kb payload is sent
                serial_payload_disable <= '1';                -- stop payload if required
					 serial_enable <= '0';
					 serial_payload_enable <= '0';
				end if;
				
			end if;
		end if;
   end process;
	
end arch_shuriken;

